@@ -1,7 +1,7 @@
 /*
  * iucode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
  *
- * Copyright (c) 2010-2014 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
+ * Copyright (c) 2010-2015 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
  *               2000 Simon Trimmer, Tigran Aivazian
  *
  * Some code copied from microcode.ctl v1.17.
@@ -28,6 +28,7 @@
 #include <argp.h>
 #include <dirent.h>
 #include <time.h>
+#include <cpuid.h>
 
 #include "intel_microcode.h"
 #include "iucode_tool_config.h"
@@ -98,23 +99,24 @@ static intel_ucode_file_type_t ucfiletype = INTEL_UC_FT_UNKNOWN;
 
 /* linked list of filenames */
 struct filename_list {
-	intel_ucode_file_type_t type;
 	struct filename_list *next;
+	intel_ucode_file_type_t type;
 	char path[];
 };
 static struct filename_list *input_files = NULL;
+static int processed_stdin = 0;
 
 /* Intel Microcode data structures */
 
 struct microcode_bundle {
 	struct microcode_bundle *next;
 	const char *filename;	/* source file name */
-	unsigned long int id;   /* bundle id */
+	unsigned int id;        /* bundle id */
 	unsigned long int size; /* bytes */
 	const void *data;       /* binary file data */
 };
 static struct microcode_bundle *microcode_bundles = NULL;
-static unsigned long int last_bundle_id = 0;
+static unsigned int next_bundle_id = 1;
 
 /* intel_uclist_entry signature flag constants */
 enum {
@@ -130,37 +132,37 @@ enum {
 
 /* signature single-linked list */
 struct intel_uclist_entry {
-	unsigned long int id;	/* microcode id within group */
-	unsigned long int gid;	/* microcode group id */
-	uint32_t flags;		/* INTEL_UCLE_ flags */
+	struct intel_uclist_entry *next;
 	uint32_t cpuid;		/* sorting key */
 	uint32_t pf_mask;
 	int32_t  uc_rev;	/* duplicated from header */
+	uint32_t flags;		/* INTEL_UCLE_ flags */
+	unsigned int id;	/* microcode id within group */
+	unsigned int gid;	/* microcode group id */
 	const void *uc;
 	uint32_t uc_size;
-	struct intel_uclist_entry *next;
 };
 static struct intel_uclist_entry *all_microcodes = NULL;
 static struct intel_uclist_entry *microcodes = NULL;
 
 /* UI helpers */
-#define UCODE_ID_MAX_LEN 20
+#define UCODE_ID_MAX_LEN 22
 struct microcode_interator_data {
 	const struct microcode_bundle *current_bundle;
 	unsigned long int total_signature_count;
 	unsigned long int total_unique_sig_count;
 	unsigned long int total_entry_count;
-	unsigned long int current_uc;
+	unsigned int current_uc;
 	char current_uc_id_str[UCODE_ID_MAX_LEN];
 };
 static struct microcode_interator_data microcode_interator_data;
 
 /* Filter masks */
 struct microcode_filter_entry {
+	struct microcode_filter_entry *next;
 	uint32_t cpuid;		/* exact match */
 	uint32_t pf_mask;	/* common bits set match */
 	int invert;
-	struct microcode_filter_entry *next;
 };
 static struct microcode_filter_entry *uc_filter_list = NULL;
 static int filter_list_allow = 1;
@@ -185,28 +187,73 @@ static unsigned int extsig_tables_in_use = 0;
 /* Helpers */
 
 #define print_msg_u(format, arg...) \
-	do { fprintf(stderr, "%s: " format "\n", progname, ## arg); } while (0)
+	do { fflush(stdout); fprintf(stderr, "%s: " format "\n", progname, ## arg); } while (0)
 
 #define print_msg(level, format, arg...) \
-	do { if (verbosity >= level) \
+	do { \
+	    if (verbosity >= level) { \
+		fflush(stdout); \
 		fprintf(stderr, "%s: " format "\n", progname, ## arg); \
+	    } \
 	} while (0)
 
 #define print_err(format, arg...) \
-	do { fprintf(stderr, "%s: " format "\n", progname, ## arg); } while (0)
+	do { fflush(stdout); fprintf(stderr, "%s: " format "\n", progname, ## arg); } while (0)
 
 static inline int is_dash(const char * const fn)
 {
 	return fn && *fn == '-' && !*(fn+1);
 }
 
+static int is_valid_fd(const int fd)
+{
+	return fcntl(fd, F_GETFD) != -1 || errno != EBADF;
+}
+
+static void fix_fds(const int fd, const int fl)
+{
+	int nfd;
+
+	if (is_valid_fd(fd))
+		return;
+
+	nfd = open("/dev/null", fl);
+	if (nfd == -1 || dup2(nfd, fd) == -1) {
+		print_err("could not attach /dev/null to file descriptor %d: %s",
+			  fd, strerror(errno));
+		exit(EXIT_SWFAILURE);
+	}
+	if (nfd != fd)
+		close(nfd);
+}
+
+/*
+ * glibc does not ensure sanity of the standard streams at program start
+ * for non suid/sgid applications.  The streams are initialized as open
+ * and not in an error state even when their underlying FDs are invalid
+ * (closed).  These FDs will later become valid due to an unrelated
+ * open(), which will cause undesired behavior (such as data corruption)
+ * should the stream be used.
+ *
+ * freopen() cannot be used to fix this directly, due to a glibc 2.14+ bug
+ * when freopen() is called on an open stream that has an invalid FD which
+ * also happens to be the first available FD.
+ */
+static void sanitize_std_fds(void)
+{
+	/* do it in file descriptor numerical order! */
+	fix_fds(STDIN_FILENO,  O_RDONLY);
+	fix_fds(STDOUT_FILENO, O_WRONLY);
+	fix_fds(STDERR_FILENO, O_RDWR);
+}
+
 /* UI microcode id */
 
-static void str_append_ucode_id(unsigned long int id,
-				unsigned long int gid,
+static void str_append_ucode_id(unsigned int id,
+				unsigned int gid,
 				char *str, unsigned int len)
 {
-	snprintf(str, len, "%02lu/%03lu", gid, id);
+	snprintf(str, len, "%02u/%03u", gid, id);
 }
 
 /* Signature linked list */
@@ -232,7 +279,7 @@ static void free_uclist(struct intel_uclist_entry *p)
 /*
  * remove entries that were superseeded
  */
-static void xx_uclist_remove_shadowed(const unsigned long int gid,
+static void xx_uclist_remove_shadowed(const unsigned int gid,
 				     const uint32_t cpuid,
 				     const uint32_t pf_mask,
 				     const int32_t uc_rev,
@@ -281,8 +328,8 @@ static void xx_uclist_remove_shadowed(const unsigned long int gid,
  *          EBADF : entry already in list, with different opaque data
  *          0:      entry not yet in list (will be inserted)
  */
-static int uclist_add_signature(const unsigned long int id,
-			const unsigned long int gid,
+static int uclist_add_signature(const unsigned int id,
+			const unsigned int gid,
 			const uint32_t flags,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
@@ -373,8 +420,8 @@ static int uclist_add_signature(const unsigned long int id,
  *                  inserted if @downgrade is non-zero)
  *          EINVAL: bad @cpuid, @uc pointer or @uclist pointer
  */
-static int uclist_merge_signature(const unsigned long int id,
-			const unsigned long int gid,
+static int uclist_merge_signature(const unsigned int id,
+			const unsigned int gid,
 			const uint32_t flags,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
@@ -501,7 +548,7 @@ static int uclist_merge_signature(const unsigned long int id,
  * Returns 0 if the bundled was added, or ENOMEM
  */
 static int add_intel_microcode_bundle(const char * const mcb_filename,
-				      unsigned long int mcb_id,
+				      unsigned int mcb_id,
 				      const void * const mcb,
 				      const size_t mcb_size)
 {
@@ -546,7 +593,8 @@ static void free_intel_microcode_bundles(void)
 	microcode_bundles = NULL;
 }
 
-static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
+static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length,
+				    unsigned long int *lineno)
 {
 	const size_t line_buffer_size = 2048;
 	const size_t buffer_size_granularity = IUCODE_TYPICAL_MCU_FILE_SIZE / sizeof(uint32_t);
@@ -558,11 +606,15 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 	uint32_t *rp;
 	size_t length;
 	size_t pos = 0;
+	unsigned long int lines = 0;
 	int err;
 
 	assert(mcb);
 	assert(mcb_length);
 
+	if (ferror(fp))
+		return -EIO;
+
 	mcb_buflen = buffer_size_granularity;
 	mcb_buffer = malloc(mcb_buflen * sizeof(uint32_t));
 	line_buffer = malloc(line_buffer_size);
@@ -574,40 +626,39 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 	err = EINVAL;
 	while (likely(fgets(line_buffer, line_buffer_size, fp) != NULL)) {
 		/*
-		 * Data lines are of the form "%x, %x, %x, %x,"
+		 * Data lines are of the form "0x%x, 0x%x, 0x%x, 0x%x,"
+		 * Comment lines start with a single "/"
 		 *
-		 * other valid lines start with / (comments), and we also
-		 * accept empty lines, and ignore whitespace.
+		 * As long as the initial comment block is stripped, the file
+		 * would be valid C source to initialize an array of 32-bit
+		 * integers.
+		 *
+		 * This parser will accept empty lines, and ignore initial
+		 * whitespace.
+		 *
+		 * It will accept a comment after valid data, as long as
+		 * there's whitespace or a comma before the comment start.
+		 *
+		 * It does not properly handle multi-line C comments
+		 * unless they start with / in every line.
+		 *
+		 * This parser is tuned to the exact format Intel has been
+		 * using for .dat files since 2008 (which are the ones that
+		 * can be distributed by Linux distros).
 		 */
 		lp = line_buffer;
+		lines++;
 		while (isspace(*lp))
 			lp++;
 
-		if (!*lp || *lp == '/')
-			continue;
-
-		while (*lp) {
+		while (likely(*lp && *lp != '/')) {
 			char *ep;
 
 			errno = 0;
 			mcb_buffer[pos] = strtoul(lp, &ep, 0);
-			if (unlikely(errno))
+			if (unlikely(errno || lp == ep))
 				goto err_out;
-			if (lp == ep) {
-				while (isspace(*lp))
-					lp++;
-				/* last value might have a , after it and
-				 * we will hit it here */
-				if (*lp == ',')
-					lp++;
-				while (isspace(*lp))
-					lp++;
-				if (unlikely(*lp))
-					goto err_out;
-				break; /* EOL */
-			}
 
-			/* we did read a value */
 			pos++;
 			if (unlikely(mcb_buflen <= pos)) {
 				/* expand buffer */
@@ -625,18 +676,23 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 				mcb_buffer = rp;
 			}
 
+			/* seek to start of next value or to EOL */
 			lp = ep;
-			/* next value */
 			while (isspace(*lp))
 				lp++;
-			if (*lp == ',') {
-				lp++;
-				continue;
-			} else if (unlikely(*lp)) {
+			if (likely(*lp == ',')) {
+				/* skip ",[[:space:]]*" */
+				do { lp++; } while (isspace(*lp));
+				/* we're either at EOL or next value, now */
+			} else if (unlikely(*lp && (lp == ep || *lp != '/'))) {
 				goto err_out;
 			}
 		}
 	}
+	if (ferror(fp)) {
+		err = -errno; /* negative! */
+		goto err_out;
+	}
 
 	length = pos * sizeof(uint32_t);
 	if (length < 1024) {
@@ -663,6 +719,9 @@ err_out:
 		*mcb_length = 0;
 	}
 
+	if (lineno)
+		*lineno = lines;
+
 	free(line_buffer);
 	return err;
 }
@@ -687,7 +746,9 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 	assert(mcb_length);
 
 	/* Try to get file size beforehand */
-	if (!fstat(fd, &stat) && S_ISREG(stat.st_mode) && stat.st_size > 0) {
+	if (fstat(fd, &stat))
+		return -errno; /* negative! */
+	if (S_ISREG(stat.st_mode) && stat.st_size > 0) {
 		if ((size_t)stat.st_size > buffer_limit)
 			return EFBIG;
 		mcb_size = stat.st_size;
@@ -782,6 +843,8 @@ static int load_intel_microcode_file(int fd,
 	void *mcb = NULL;
 	size_t mcb_length = 0;
 
+	unsigned long int tl = 0;
+
 	switch (ftype) {
 	case INTEL_UC_FT_BIN:
 		err = load_intel_microcode_bin(fd, &mcb, &mcb_length);
@@ -792,7 +855,7 @@ static int load_intel_microcode_file(int fd,
 			fp = fdopen(fd, "r");
 
 		if (fp) {
-			err = load_intel_microcode_dat(fp, &mcb, &mcb_length);
+			err = load_intel_microcode_dat(fp, &mcb, &mcb_length, &tl);
 		} else {
 			err = -errno;
 		}
@@ -808,8 +871,7 @@ static int load_intel_microcode_file(int fd,
 	}
 
 	if (!err && mcb) {
-		last_bundle_id++;
-		err = add_intel_microcode_bundle(fn, last_bundle_id,
+		err = add_intel_microcode_bundle(fn, next_bundle_id,
 						 mcb, mcb_length);
 		if (!err)
 			mcb = NULL; /* someone else will free it */
@@ -817,14 +879,19 @@ static int load_intel_microcode_file(int fd,
 
 	switch (err) {
 	case 0:
-		print_msg(2, "microcode bundle %lu: %s (%zu bytes)",
-			  last_bundle_id, fn, mcb_length);
+		print_msg(2, "microcode bundle %u: %s (%zu bytes)",
+			  next_bundle_id, fn, mcb_length);
+		next_bundle_id++;
 		break;
 	case ENOMEM:
 		print_err("%s: could not allocate memory while loading", fn);
 		break;
 	case EINVAL:
-		print_err("%s: invalid file format", fn);
+		if (!tl) {
+			print_err("%s: invalid file format", fn);
+		} else {
+			print_err("%s: line %lu: invalid file format", fn, tl);
+		}
 		break;
 	case EFBIG:
 		print_err("%s: cowardly refusing to load an insanely large data file", fn);
@@ -858,13 +925,19 @@ static int load_intel_microcode(const char * path,
 
 	/* Should we read from stdin ? */
 	if (is_dash(path)) {
+		/* read stdin only once, ignore further requests */
+		if (processed_stdin)
+			return 0;
+
+		processed_stdin = 1;
 		ftype = baseftype;
 
 		/* default to .dat mode for stdin */
 		if (ftype == INTEL_UC_FT_UNKNOWN)
 			ftype = INTEL_UC_FT_DAT;
 
-		return load_intel_microcode_file(0, stdin, "(stdin)", ftype);
+		return load_intel_microcode_file(fileno(stdin), stdin,
+						"(stdin)", ftype);
 	}
 
 	dir = NULL;
@@ -1377,6 +1450,7 @@ static int is_in_date_range(const struct intel_ucode_metadata * const m)
 	return !!((d > datefilter_min) && (d < datefilter_max));
 }
 
+static void free_filter_list(struct microcode_filter_entry *f) __attribute__((unused));
 static void free_filter_list(struct microcode_filter_entry *f)
 {
 	static struct microcode_filter_entry *p;
@@ -1570,7 +1644,7 @@ static int xx_uclist_add_print_errors(int status)
 		print_err("Cannot add index entry: out of memory");
 		return 1;
 	case EINVAL:
-		print_err("Internal error uclist_merge_signature() returned EINVAL");
+		print_err("Internal error: uclist_merge_signature() returned EINVAL");
 		return 1;
 	default:
 		return 1;
@@ -1675,6 +1749,9 @@ static int xx_process_ucode_signature_cb(void *userdata,
 			  "and signature as a previously loaded microcode, "
 			  "but different contents", ctx->current_uc_id_str);
 		return 1;
+	case EINVAL:
+		print_err("Internal error: uclist_add_signature() returned EINVAL");
+		return 1;
 	default:
 		print_err("Failed to add microcode entry %s: %s",
 			  ctx->current_uc_id_str, strerror(add_status));
@@ -1741,10 +1818,10 @@ static int do_process_microcodes(void)
 	while (mcb) {
 		if (list_all_microcodes) {
 			if (mcb->filename) {
-				printf("microcode bundle %lu: %s\n",
+				printf("microcode bundle %u: %s\n",
 					mcb->id, mcb->filename);
 			} else {
-				printf("microcode bundle %lu:\n",
+				printf("microcode bundle %u:\n",
 					mcb->id);
 			}
 		}
@@ -1779,8 +1856,8 @@ static int do_process_microcodes(void)
 	if (list_sel_microcodes || datefilter_loose || verbosity >= 2) {
 		struct intel_uclist_entry *uce = microcodes;
 		struct intel_uclist_entry *ucl = all_microcodes;
-		unsigned long uccount = 0;
-		unsigned long sigcount = 0;
+		unsigned long int uccount = 0;
+		unsigned long int sigcount = 0;
 
 		if (list_sel_microcodes)
 			printf("selected microcodes:\n");
@@ -2008,16 +2085,42 @@ err_exit:
 	return rc;
 }
 
-static int scan_system_processors(void)
+static int xx_scan_handle_sig(uint32_t const id1, uint32_t const id2,
+				uint32_t const id3, uint32_t const sig,
+				struct microcode_filter_entry ** const ucfp)
 {
-	char cpuid_device[PATH_MAX];
-	uint32_t cpuid_buf[8];
-	struct microcode_filter_entry *uc_cpu = NULL;
+	/* Is it a supported Intel processor ? */
+	if (id1 == 0x756e6547 && /* "Genu" */
+	    id3 == 0x49656e69 && /* "ineI" */
+	    id2 == 0x6c65746e) { /* "ntel" */
+		/*
+		 * Get main processor signature from cpuid(1) EAX
+		 * and add it as a filter.
+		 */
+		switch (add_filter_to_list(sig, 0, 0, ucfp)) {
+		case ENOMEM:
+			print_err("out of memory");
+			return ENOMEM;
+		case EEXIST:
+			return 0;
+		default:
+			print_msg(1, "system has processor(s) with signature 0x%08x", sig);
+			return 0;
+		}
+	}
 
+	return ENXIO;
+}
+
+#ifdef USE_CPUID_DEVICE
+static int check_cpuid_devs(struct microcode_filter_entry ** const ucfp)
+{
+	uint32_t cpuid_buf[8];	/* two cpuid levels */
+	char cpuid_device[PATH_MAX];
 	int cpuid_fd;
+	int rc = 0;
 	unsigned int i = 0;
 	unsigned int ncpu = 0;
-	int rc = 0;
 
 	while (1) {
 		snprintf(cpuid_device, sizeof(cpuid_device),
@@ -2050,82 +2153,88 @@ static int scan_system_processors(void)
 		print_msg(3, "trying to get CPUID information from %s",
 			  cpuid_device);
 		if (read(cpuid_fd, &cpuid_buf, sizeof(cpuid_buf)) == -1) {
-			print_err("%s: access to CPUID(0) and CPUID(1) failed",
-				  cpuid_device);
+			print_err("%s: access to CPUID(0) and CPUID(1) failed: %s",
+				  cpuid_device, strerror(errno));
 			/* this is in the should not happen list, so abort */
 			close(cpuid_fd);
-			rc = 1;
-			goto err_out;
+			return 1;
 		}
 
 		close(cpuid_fd);
 		ncpu++;
 
-		/* Is it a supported Intel processor ? */
-		if (cpuid_buf[0] > 0 && /* we need cpuid(1) to exist */
-		    cpuid_buf[1] == 0x756e6547 && /* "Genu" */
-		    cpuid_buf[3] == 0x49656e69 && /* "ineI" */
-		    cpuid_buf[2] == 0x6c65746e) { /* "ntel" */
-			/*
-			 * Get main processor signature from cpuid(1) EAX
-			 * and add it as a filter.
-			 */
-			switch (add_filter_to_list(cpuid_buf[4], 0, 0, &uc_cpu)) {
-			case ENOMEM:
-				print_err("out of memory");
-				rc = 1;
-				goto err_out;
-			case EEXIST:
-				break;
-			default:
-				print_msg(1, "system has processor(s) with signature 0x%08x",
-					  cpuid_buf[4]);
-			}
-		} else {
-			if (!uc_cpu) {
-				/* shortcut search, assume none will be found */
-				print_msg(2, "non-Intel processor found, skipping scan");
-				break;
-			}
+		if (xx_scan_handle_sig(cpuid_buf[1], cpuid_buf[2],
+				       cpuid_buf[3], cpuid_buf[4], ucfp) == ENOMEM) {
+			rc = -1;
+			break;
 		}
 
 		i++;
 	};
 
-	/*
-	 * Be helpful: tell user why we're not doing what he explicitly
-	 * requested if we cannot get any cpuids at all.
-	 *
-	 * fail-safe: we only switch away from "select all microcodes by default"
-	 * if we managed to scan, so we will include all microcodes if cpuid is not
-	 * available, and no other microcode selection option was used
-	 */
 	if (i == 0 && ncpu == 0) {
-		print_err("cpuid kernel driver unavailable, cannot scan system processor signatures");
-		rc = 0;
-		goto err_out;
+		print_msg(2, "cpuid kernel driver unavailable");
+		return 0;
 	} else if (rc) {
 		if (ncpu)
 			print_msg(1, "some processors were not scanned due to unexpected errors");
 		else
-			print_err("could not open cpuid devices, cannot scan system processor signatures");
+			print_msg(2, "could not open cpuid devices");
 	}
 
-	rc = 0;
+	if (ncpu)
+		print_msg(2, "checked the signature of %u processor(s)", ncpu);
 
-	/* we managed to scan, do not select every microcode by default */
-	filter_list_allow = 0;
+	return 0;
+}
+#else
+static int check_cpuid_devs(__attribute__((unused)) struct microcode_filter_entry ** const uc_cpu)
+{
+	print_msg(2, "assuming all processors have the same signature");
+	return 0;
+}
+#endif
+
+static int scan_system_processors(void)
+{
+	uint32_t id0, id1, id2, id3, sig, idx;
+	struct microcode_filter_entry *uc_cpu = NULL;
+	int rc = 0;
+
+	print_msg(3, "trying to get CPUID information directly");
+	if (!(__get_cpuid(0, &id0, &id1, &id2, &id3) &&
+	      __get_cpuid(1, &sig, &idx, &idx, &idx))) {
+		print_msg(1, "microcode signature unavailable");
+		return 0;
+	}
+
+	/*
+	 * fail-safe: only change filter_list_allow (switch away from "select
+	 * all microcodes by default") if we did scan/cpuid.  This way, all
+	 * microcodes will be included if cpuid is not available, and no other
+	 * microcode selection option was used.  On non-Intel, this results in
+	 * "no microcodes by default", because the scan/cpuid was sucessful,
+	 * but uc_cpu will be empty.
+	 */
+
+	switch (xx_scan_handle_sig(id1, id2, id3, sig, &uc_cpu)) {
+	case 0:
+		filter_list_allow = 0;
+		rc = check_cpuid_devs(&uc_cpu);
+		break;
+	case ENXIO:
+		filter_list_allow = 0;
+		print_msg(2, "running on a non-Intel processor");
+		break;
+	default:
+		rc = 1;
+	}
 
 	if (uc_cpu) {
 		/* tie linked lists */
 		add_filter_list_to_list(&uc_filter_list, uc_cpu);
 		uc_cpu = NULL;
 	}
-	print_msg(2, "checked the signature of %u processor(s)", ncpu);
-
-err_out:
-	if (uc_cpu)
-		free_filter_list(uc_cpu);
 
 	return rc;
 }
@@ -2134,7 +2243,7 @@ err_out:
 
 static const char program_version[] =
 	PROGNAME " " VERSION "\n"
-	"Copyright (c) 2010-2014 by Henrique de Moraes Holschuh\n\n"
+	"Copyright (c) 2010-2015 by Henrique de Moraes Holschuh\n\n"
 
 	"Based on code from the Linux microcode_intel driver and from\n"
 	"the microcode.ctl package, copyright (c) 2000 by Simon Trimmer\n"
@@ -2212,27 +2321,27 @@ static struct argp_option cmdline_options[] = {
 	  20 },
 
 	{ "downgrade", IUCODE_ARGP_DOWNGRADE, NULL, 0,
-	  "Instead of discarding microcodes based on revision level, "
-	  "keep the one from the file loaded last.  Files are loaded "
-	  "in the order they were specified in the command line",
+	   "Instead of discarding microcodes based on revision level, "
+	   "keep the one from the file loaded last.  Files are loaded "
+	   "in the order they were specified in the command line",
 	  25 },
 	{ "no-downgrade", IUCODE_ARGP_NODOWNGRADE, NULL, 0,
-	  "Keep the microcode with the highest revision level, regardless "
-	  "of the file load order (default)",
+	   "Keep the microcode with the highest revision level, regardless "
+	   "of the file load order (default)",
 	  25 },
 
 	{ "date-before", IUCODE_ARGP_DATEBEFORE, "YYYY-MM-DD", 0,
-	  "Select only microcodes older than the specified date",
+	   "Select only microcodes older than the specified date",
 	  27 },
 	{ "date-after", IUCODE_ARGP_DATEAFTER, "YYYY-MM-DD", 0,
-	  "Select only microcodes newer than the specified date",
+	   "Select only microcodes newer than the specified date",
 	  27 },
 	{ "loose-date-filtering", IUCODE_ARGP_DATEFLOOSE, NULL, 0,
-	  "Consider for selection other revisions (outside of the date range) "
-	  "of every microcode that was selected within the date range",
+	   "Consider for selection other revisions (outside of the date range) "
+	   "of every microcode that was selected within the date range",
 	  28 },
 	{ "strict-date-filtering", IUCODE_ARGP_DATEFSTRICT, NULL, 0,
-	  "Select only microcodes strictly within the date range (default)",
+	   "Select only microcodes strictly within the date range (default)",
 	  28 },
 
 	{ "list",     'l', NULL, 0, "List selected microcode signatures", 30 },
@@ -2253,13 +2362,13 @@ static struct argp_option cmdline_options[] = {
 	   "The binary format is suitable to be uploaded to the kernel",
 	  40 },
 	{ "write-named-to", 'W', "directory", 0,
-	  "Write selected microcodes to files in the specified directory, "
-	  "in binary format.  The file name will reflect the microcode "
-	  "signature, mask and revision",
+	   "Write selected microcodes to files in the specified directory, "
+	   "in binary format.  The file name will reflect the microcode "
+	   "signature, mask and revision",
 	  40 },
 	{ "write-earlyfw", IUCODE_ARGP_EIRFS, "file", 0,
-	  "Write selected microcodes to an early initramfs file, which "
-	  "should be prepended to the regular initramfs",
+	   "Write selected microcodes to an early initramfs file, which "
+	   "should be prepended to the regular initramfs",
 	  40 },
 
 	{ "overwrite", IUCODE_ARGP_UNLINK, NULL, 0,
@@ -2277,10 +2386,10 @@ static struct argp_option cmdline_options[] = {
 	  51 },
 
 	{ "ignore-broken", IUCODE_ARGP_IGNOREBROKEN, NULL, 0,
-	  "Skip broken microcode entries instead of aborting",
+	   "Skip broken microcode entries instead of aborting",
 	  55 },
 	{ "no-ignore-broken", IUCODE_ARGP_NOIGNOREBROKEN, NULL, 0,
-	  "Abort on broken microcode entries (default)",
+	   "Abort on broken microcode entries (default)",
 	  56 },
 
 	{ 0 },
@@ -2396,9 +2505,6 @@ static int cmdline_get_date(const char *arg, uint32_t *d)
 	unsigned long int year, month, day;
 	char *p;
 
-	while (isspace(*arg))
-		arg++;
-
 	errno = 0;
 	year = strtoul(arg, &p, 10);
 	if (errno || p == arg || *p != '-')
@@ -2458,7 +2564,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 
 	case 't':
 		if (strlen(arg) > 1)
-			argp_error(state, "unknown file type: %s\n", arg);
+			argp_error(state, "unknown file type: '%s'", arg);
 		switch (*arg) {
 		case 'd': /* .dat */
 			ucfiletype = INTEL_UC_FT_DAT;
@@ -2470,14 +2576,14 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 			ucfiletype = INTEL_UC_FT_UNKNOWN;
 			break;
 		default:
-			argp_error(state, "unknown file type: %c\n", *arg);
+			argp_error(state, "unknown file type: '%c'", *arg);
 		}
 		break;
 
 	case 'k':
 		if (command_line_actions & IUCODE_DO_UPLOADUC)
 			argp_error(state,
-				   "-k option can be specified only once\n");
+				   "-k option can be specified only once");
 
 		if (arg)
 			upload_microcode = strdup(arg);
@@ -2489,7 +2595,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	case 'K':
 		if (command_line_actions & IUCODE_DO_WRITEFW)
 			argp_error(state,
-				   "-K option can be specified only once\n");
+				   "-K option can be specified only once");
 
 		if (arg)
 			write_firmware = strdup(arg);
@@ -2501,7 +2607,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	case 'w':
 		if (command_line_actions & IUCODE_DO_WRITEUC)
 			argp_error(state,
-				   "-w option can be specified only once\n");
+				   "-w option can be specified only once");
 
 		write_microcode = strdup(arg);
 		command_line_actions |= IUCODE_DO_WRITEUC;
@@ -2509,7 +2615,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	case IUCODE_ARGP_EIRFS:
 		if (command_line_actions & IUCODE_DO_WRITEFWE)
 			argp_error(state,
-				   "--write-earlyfw option can be specified only once\n");
+				   "--write-earlyfw option can be specified only once");
 
 		write_early_firmware = strdup(arg);
 		command_line_actions |= IUCODE_DO_WRITEFWE;
@@ -2517,7 +2623,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	case 'W':
 		if (command_line_actions & IUCODE_DO_WRITEFWN)
 			argp_error(state,
-				   "-W option can be specified only once\n");
+				   "-W option can be specified only once");
 
 		write_named = strdup(arg);
 
@@ -2538,7 +2644,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 		case EEXIST:
 			break; /* success */
 		case EINVAL:
-			argp_error(state, "invalid filter: %s", arg);
+			argp_error(state, "invalid filter: '%s'", arg);
 			break; /* not reached */
 		default:
 			argp_failure(state, EXIT_SWFAILURE, rc,
@@ -2554,7 +2660,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	case IUCODE_ARGP_DATEAFTER:
 		if (cmdline_get_date(arg, (key == IUCODE_ARGP_DATEBEFORE) ?
 					   &datefilter_max : &datefilter_min))
-			argp_error(state, "invalid date: %s", arg);
+			argp_error(state, "invalid date: '%s'", arg);
 		command_line_actions |= IUCODE_F_UCSELECT;
 		break;
 	case IUCODE_ARGP_DATEFSTRICT:
@@ -2610,6 +2716,8 @@ int main(int argc, char *argv[])
 
 	progname = argv[0];
 
+	sanitize_std_fds();
+
 	argp_err_exit_status = EXIT_USAGE;
 	argp_program_version = program_version;
 	argp_program_bug_address = NULL; /* FIXME */
@@ -2628,7 +2736,7 @@ int main(int argc, char *argv[])
 
 	if (command_line_actions & IUCODE_DO_LOADFILE) {
 		struct filename_list *fn = input_files;
-		while (fn) {
+		while (fn && next_bundle_id) {
 			switch (load_intel_microcode(fn->path, fn->type)) {
 			case 0:
 				break;
@@ -2643,6 +2751,12 @@ int main(int argc, char *argv[])
 			}
 			fn = fn->next;
 		}
+		if (!next_bundle_id) {
+			/* too many bundles: gid overflow */
+			print_err("too many data files");
+			rc = EXIT_SWFAILURE;
+			goto err_exit;
+		}
 
 		if (microcode_bundles && do_process_microcodes()) {
 			rc = EXIT_SWFAILURE;
@@ -2650,6 +2764,9 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	/* flush --list-* output if it is still pending */
+	fflush(stdout);
+
 	if (command_line_actions & IUCODE_DOMASK_NEEDSUC) {
 		if (microcodes) {
 			rc = uclist_annotate_extsig_dup(microcodes);
